from __future__ import print_function

from ghizmo.commands import lib

import os
import codecs
import yaml
import urllib
from collections import defaultdict

_ROLES_FILES = ["authors-info.yml", "authors-info.json", "admin/authors-info.yml", "admin/authors-info.json"]

def assemble_authors(config, args):
  """
  Assemble a list of authors as an AUTHORS.md file based on GitHub repo history and a
  authors-info.{yml,json} file.
  """
  github = config.github
  repo = config.repo
  roles_filename = None
  for filename in _ROLES_FILES:
    if os.path.isfile(filename):
      roles_filename = filename

  header = None
  footer = None
  roles = {}
  if roles_filename:
    yield lib.status("Info from: %s" % roles_filename)
    with codecs.open(roles_filename, "r", "utf-8") as f:
      info = yaml.safe_load(f)
      header = info.get("header")
      footer = info.get("footer")
      roles = info.get("roles")
      exclude = info.get("exclude")
  else:
    yield lib.status("No roles file")

  author_list = []
  for contributor in repo.contributors():
    user = github.user(contributor.login)
    author_list.append((user.login, user.name, roles.get(user.login)))

  yield lib.status("Read %s authors" % len(author_list))

  # Sort alphabetically by login.
  author_list.sort(key=lambda (login, name, role): login.lower())

  def format_user(login, name):
    if login and name:
      return u"%s (%s)" % (name, login)
    elif login:
      return login
    else:
      raise ValueError("Missing login name")

  commit_tallies = {}
  for stat in repo.contributor_statistics():
    commit_tallies[stat.author.login] = stat.total

  yield lib.status("Read %s contributor stats" % len(commit_tallies))

  issue_tallies = defaultdict(int)
  for issue in repo.issues(state="all"):
    issue_tallies[issue.user.login] += 1

  yield lib.status("Read %s issues/PRs" % len(issue_tallies))

  with codecs.open("AUTHORS.md", "w", "utf-8") as f:
    f.write(u"# Authors\n\n")
    if header:
      f.write(u"%s\n\n" % header)
    for (login, name, role) in author_list:
      if login in exclude:
        continue

      user_url = "https://github.com/%s" % login
      # Link to commits by that author
      commits_count = commit_tallies.get(login, 0)
      commits_url = "%s/commits?author=%s" % (repo.html_url, urllib.quote_plus(login))
      # Link to issues and PRs by that author.
      issues_count = issue_tallies.get(login, 0)
      issues_url = "%s/issues?q=%s" % (repo.html_url, urllib.quote_plus("author:%s" % login))

      role_str = u" \u2014 _%s_" % role if role else ""
      f.write(u"* [%s](%s) \u2014 [%s+](%s)/[%s+](%s)%s\n" % \
        (format_user(login, name), user_url, commits_count, commits_url, issues_count, issues_url, role_str))
    if footer:
      f.write(u"\n%s\n\n" % footer)

    f.write(u"\n(This file was auto-generated by [ghizmo assemble-authors](https://github.com/jlevy/ghizmo).)")



"""
Ghizmo main library.
"""

from __future__ import print_function

__author__ = 'jlevy'

import logging as log
import sys
import re
import json
import yaml
import inspect
import os
import importlib
import getpass
from collections import namedtuple
from collections import OrderedDict
from functools32 import lru_cache  # functools32 pip
import github3  # github3.py pip
from github3.null import NullObject

import configs
import commands


def read_login_info(username=None):
  if not username:
    username = raw_input("GitHub username: ")
  return (username, getpass.getpass())


def login(username=None):
  token = configs.get_access_token()
  if token:
    log.info("Using access token authentication")
    return github3.login(token=token)
  else:
    username = username or configs.get_username()
    return github3.login(*read_login_info(username=username))


def format_to_string(obj, format=None):
  if not format:
    format = "json"

  if isinstance(obj, NullObject):
    raise ValueError("Command returned null type (invalid input?): %s" % repr(obj))

  # Heuristic to convert github3 objects to serializable form.
  if hasattr(obj, "as_dict"):
    obj = obj.as_dict()

  if format == "json":
    return json.dumps(obj, sort_keys=True, indent=2) + "\n"
  elif format == "yaml":
    return yaml.safe_dump(obj, default_style='"')
  else:
    raise AssertionError("Invalid format: %s" % format)


def print_formatter(format=None):
  def printer(obj):
    print(format_to_string(obj, format), end="")
    sys.stdout.flush()

  return printer


# All data used by the run of a command.
Config = namedtuple("Config", "github repo formatter")


def _to_dash(name):
  return name.replace("_", "-")


def _to_underscore(name):
  return name.replace("-", "_")


def _is_public_func(f):
  return inspect.isfunction(f) and not f.__name__.startswith("_")


@lru_cache()
def all_command_functions():
  # TODO: Clean this up somehow.
  # Use absolute module path so it works post installation.
  modules = [importlib.import_module("ghizmo.commands.%s" % module) for module in commands.__all__]

  # If there is a ghizmo_commands.py file in the current directory, use it too.
  if os.path.exists("ghizmo_commands.py"):
    sys.path.insert(1, '.')  # This is needed only on some installations.
    modules.append(importlib.import_module("ghizmo_commands"))

  log.info("Imported command modules: %s", modules)

  func_map = {}
  for module in modules:
    for (name, func) in inspect.getmembers(module, predicate=_is_public_func):
      if name in func_map:
        raise ValueError("Duplicate function name for command: %s" % name)
      func_map[name] = func
  # Sort items first by module, then by name.
  return OrderedDict(sorted(func_map.items(), key=lambda (name, func): (func.__module__, name)))


@lru_cache()
def command_directory(use_dashes=True):
  def doc_for_function(func):
    doc = func.__doc__ and re.sub("\s+", " ", func.__doc__).strip()
    if not doc:
      doc = "(no pydoc)"
    return doc

  transform = _to_dash if use_dashes else lambda x: x
  return [(func.__module__, transform(name), doc_for_function(func)) for (name, func) in
          all_command_functions().iteritems()]


@lru_cache()
def list_commands(use_dashes=True):
  return [command for (module, command, doc) in command_directory(use_dashes=use_dashes)]


def get_command_func(command):
  command = _to_underscore(command)
  if not command in all_command_functions():
    raise ValueError("invalid command: %s" % command)
  return all_command_functions()[command]


def run_command(command, config, args):
  command_func = get_command_func(command)
  log.info("Command '%s' (%s)", command, command_func)
  log.info("Config: %s", config)
  log.info("Args: %s", args)
  # This executes the command step by step, either just to display results, or to display progress
  # on an action with side effects.
  iterable_result = command_func(config, args)
  if iterable_result:
    for result in iterable_result:
      config.formatter(result)

# TODO:
# Control pretty-printing, using one object per line by default, but with --pretty option to print nicely
# Proper reading of JSON streams (as opposed to line-by-line), for use in piping
# Prettier SIGPIPE handling



#!/usr/bin/env python
"""
If commands require input, it must be line-delimited JSON (e.g. quoted strings).

For further documentation, see: https://github.com/jlevy/ghizmo
"""

from __future__ import print_function

__author__ = 'jlevy'

import logging as log
import sys
import os
import argparse
from commands.lib import to_bool

NAME = "ghizmo"
VERSION = "0.1.9"
DESCRIPTION = "ghizmo: An extensible command line for GitHub"
LONG_DESCRIPTION = __doc__


def log_setup(level):
  if level == log.DEBUG:
    log.basicConfig(format="%(levelname).1s %(filename)20s:%(lineno)-4d  %(message)s", level=level,
                    stream=sys.stderr)
  else:
    log.basicConfig(format="%(message)s", level=level, stream=sys.stderr)

    def brief_excepthook(exctype, value, traceback):
      print("error: %s" % value, file=sys.stderr)
      print("(run with --debug for traceback info)", file=sys.stderr)
      sys.exit(2)

    sys.excepthook = brief_excepthook


class UserArgs(object):
  """
  Assembles user-supplied arguments plus environment vars for convenient access.
  """

  def __init__(self):
    self._explicit_keys = []
    self.dict = {}

  def add_explicit(self, d):
    self.dict.update(d)
    self._explicit_keys.extend(d.keys())

  def add_implicit(self, d):
    self.dict.update(d)

  def get(self, item, default=None):
    return self.dict[item] if item in self.dict else default

  def get_bool(self, item, default=False):
    return to_bool(self.get(item, default))

  def __getattr__(self, item):
    try:
      return self.dict[item]
    except KeyError:
      raise AttributeError("Missing user-supplied argument '%s' (set with: --arg %s=VALUE)" % (item, item))

  def __str__(self):
    return ", ".join(["%s=%s" % (key, self.dict[key]) for key in self._explicit_keys])


def assemble_args(cmdline_args):
  assembled = UserArgs()

  user_arg_list = cmdline_args.arg or []
  d = {}
  for arg in user_arg_list:
    try:
      (key, value) = arg.split("=", 1)
    except:
      raise ValueError("Could not parse argument -- invalid format: '%s'" % arg)
    d[key] = value

  # Arguments are explicit, from command line, and from environment.
  assembled.add_implicit(os.environ)
  assembled.add_explicit(d)
  assembled.add_explicit({
    "dry_run": cmdline_args.dry_run,
    "force": cmdline_args.force,
    "format": cmdline_args.format
  })

  return assembled


def main():
  # Bootstrap logging right up front, so we do it before assembling commands for help.
  log_setup(log.DEBUG if "--debug" in sys.argv else log.WARN)

  import ghizmo
  import configs

  command_directory = ghizmo.command_directory()
  command_modules = sorted(set([module for (module, name, doc) in command_directory]))
  command_docs = \
    "command modules: %s\n" % ", ".join(command_modules) \
    + "(augment by adding to ./ghizmo_commands.py)\n\n" \
    + "commands available:\n" \
    + "\n".join(["  %s: %s" % (name, doc) for (module, name, doc) in command_directory])
  parser = argparse.ArgumentParser(description=DESCRIPTION, version=VERSION,
                                   epilog="\n" + __doc__ + "\n" + command_docs,
                                   formatter_class=argparse.RawTextHelpFormatter)
  parser.add_argument("command", help="%s command" % NAME, choices=ghizmo.list_commands())
  parser.add_argument("--username", help="username to log in as")
  parser.add_argument("--repo", help="repo of the form: owner/repo-name")
  parser.add_argument("--debug", help="enable debugging output", action="store_true")

  # Command arguments:
  parser.add_argument("-f", "--force", help="enable debugging output", action="store_true")
  parser.add_argument("-n", "--dry-run", help="dry run: log actions but don't do anything", action="store_true")
  parser.add_argument("--format", help="output format", choices=["json", "yaml"])
  parser.add_argument("-a", "--arg", help="argument of the form key=value (may repeat this)", action="append")

  args = parser.parse_args()

  # Validate credentials and log in.
  gh = ghizmo.login(username=args.username)
  if not gh:
    raise ValueError("Login failure")

  # Validate repository.
  owner = None
  repo_name = None
  if args.repo:
    try:
      (owner, repo_name) = args.repo.split("/")
    except:
      raise ValueError("Invalid repository (use format owner/repo-name): %s" % args.repo)
  else:
    try:
      (owner, repo_name) = configs.infer_repo()
    except:
      log.debug("couldn't infer repository", exc_info=True)

  repo = None
  if owner and repo_name:
    repo = gh.repository(owner, repo_name)
    if not repo:
      raise ValueError("Couldn't access repository: %s/%s" % (owner, repo_name))

  # Assemble config for this run.
  formatter = ghizmo.print_formatter(args.format)
  config = ghizmo.Config(github=gh, repo=repo, formatter=formatter)

  ghizmo.run_command(args.command, config, assemble_args(args))


if __name__ == '__main__':
  main()
